"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
var DirectiveType;
(function (DirectiveType) {
    DirectiveType[DirectiveType["Begin"] = 0] = "Begin";
    DirectiveType[DirectiveType["End"] = 1] = "End";
    DirectiveType[DirectiveType["Assign"] = 2] = "Assign";
    DirectiveType[DirectiveType["Merge"] = 3] = "Merge";
    DirectiveType[DirectiveType["Emit"] = 4] = "Emit";
})(DirectiveType || (DirectiveType = {}));
class InvalidSourceError extends Error {
    constructor(message, position = null) {
        super(message);
        this.position = position;
    }
    setPosition(position) {
        this.position = position;
        this.message = `${positionToString(this.position)} ${this.message}`;
    }
}
exports.InvalidSourceError = InvalidSourceError;
class InvalidDirectiveError extends InvalidSourceError {
}
exports.InvalidDirectiveError = InvalidDirectiveError;
class InvalidRegionError extends InvalidSourceError {
}
exports.InvalidRegionError = InvalidRegionError;
function createDirectiveMatcher(prefix) {
    return new RegExp(`^[ \t]*(\/\/\\s+${prefix}:(.+))$`, "gm");
}
exports.createDirectiveMatcher = createDirectiveMatcher;
function extractRegions(text, directiveMatcher, data = {}) {
    const regions = [];
    enterScope(text, regions, extractDirectives(directiveMatcher, text), 0, 0, data);
    return regions;
}
exports.extractRegions = extractRegions;
function inject(text, directiveMatcher, cb, data = {}) {
    let r = "";
    let start = 0;
    for (const region of extractRegions(text, directiveMatcher)) {
        r += text.substring(start, region.start);
        r += ensureStringStartsWithNewline(ensureStringEndsWithNewline(cb(region)));
        start = region.end;
    }
    return r += text.substring(start);
}
exports.inject = inject;
function directiveTypeFromString(s) {
    switch (s) {
        case "begin":
            return DirectiveType.Begin;
        case "end":
            return DirectiveType.End;
        case "assign":
            return DirectiveType.Assign;
        case "merge":
            return DirectiveType.Merge;
        case "emit":
            return DirectiveType.Emit;
    }
    throw new InvalidDirectiveError(`Invalid directive type: ${s}.`);
}
function parseDirective(s) {
    const pStart = s.indexOf("(");
    if (pStart !== -1) {
        const type = directiveTypeFromString(s.substring(0, pStart));
        switch (type) {
            case DirectiveType.Begin:
            case DirectiveType.End:
                throw new InvalidDirectiveError(`Invalid directive. ${DirectiveType[type]} directive should not have arguments.`);
        }
        const pEnd = s.lastIndexOf(")");
        if (pEnd === -1) {
            throw new InvalidDirectiveError(`Invalid directive. Unable to find closing parenthesis in a directive "${s}".`);
        }
        const sArg = s.substring(pStart + 1, pEnd);
        let arg;
        switch (type) {
            case DirectiveType.Assign:
            case DirectiveType.Merge:
                try {
                    arg = JSON.parse(sArg);
                }
                catch (_a) {
                    throw new InvalidDirectiveError(`Invalid directive. Unable to parse(JSON.parse) directive argument "${sArg}".`);
                }
                if (typeof arg !== "object" || arg === null) {
                    throw new InvalidDirectiveError(`Invalid ${DirectiveType[type]} directive. Argument should have an object type.`);
                }
                break;
            case DirectiveType.Emit:
                try {
                    arg = JSON.parse(`[${sArg}]`);
                }
                catch (_b) {
                    throw new InvalidDirectiveError(`Invalid directive. Unable to parse(JSON.parse) directive argument "[${sArg}]".`);
                }
                break;
        }
        return { type, arg };
    }
    else {
        const type = directiveTypeFromString(s);
        switch (type) {
            case DirectiveType.Assign:
            case DirectiveType.Merge:
            case DirectiveType.Emit:
                throw new InvalidDirectiveError(`Invalid directive. ${DirectiveType[type]} directive should have arguments.`);
        }
        return { type, arg: null };
    }
}
function extractDirectives(directiveMatcher, text) {
    const directives = [];
    let match;
    while ((match = directiveMatcher.exec(text)) !== null) {
        const fullMatch = match[0];
        const start = match.index;
        const end = start + fullMatch.length;
        const comment = match[1];
        let directive;
        try {
            directive = parseDirective(match[2]);
        }
        catch (e) {
            if (e instanceof InvalidSourceError) {
                e.setPosition(positionFromOffset(text, start));
            }
            throw e;
        }
        const padding = fullMatch.substring(0, fullMatch.length - comment.length);
        directives.push({
            type: directive.type,
            arg: directive.arg,
            padding,
            start,
            end,
        });
    }
    return directives;
}
function enterEmit(text, regions, directives, index, data, args, padding, start) {
    while (index < directives.length) {
        const directive = directives[index++];
        switch (directive.type) {
            case DirectiveType.End:
                regions.push({ args, data, padding, start, end: directive.start });
                return index;
            default:
                throw new InvalidRegionError("Emit region should not contain any directives", positionFromOffset(text, directive.start));
        }
    }
    throw new InvalidRegionError("Emit region should end with End directive", positionFromOffset(text, start));
}
function enterScope(text, regions, directives, index, scopes, data) {
    while (index < directives.length) {
        const directive = directives[index++];
        switch (directive.type) {
            case DirectiveType.Begin:
                index = enterScope(text, regions, directives, index, scopes + 1, data);
                break;
            case DirectiveType.End:
                return index;
            case DirectiveType.Assign:
                data = Object.assign({}, data, directive.arg);
                break;
            case DirectiveType.Merge:
                data = lodash_1.merge({}, data, directive.arg);
                break;
            case DirectiveType.Emit:
                index = enterEmit(text, regions, directives, index, data, directive.arg, directive.padding, directive.end);
                break;
        }
    }
    if (scopes > 0) {
        throw new InvalidRegionError(`All scopes should end with End directive.`);
    }
    return index;
}
function positionFromOffset(s, offset) {
    if (offset > s.length) {
        throw new Error("Invalid offset. Offset is pointing outside of text.");
    }
    let line = 1;
    let col = 1;
    for (let i = 0; i < offset; i++) {
        const c = s.charCodeAt(i);
        if (c === 10) {
            line++;
            col = 1;
        }
        else {
            col++;
        }
    }
    return { line, col };
}
function positionToString(pos) {
    return `[${pos.line}:${pos.col}]`;
}
function ensureStringStartsWithNewline(s) {
    if (s.length === 0) {
        return "\n";
    }
    if (s.charCodeAt(0) !== 10) {
        return "\n" + s;
    }
    return s;
}
function ensureStringEndsWithNewline(s) {
    if (s.length === 0) {
        return "\n";
    }
    if (s.charCodeAt(s.length - 1) !== 10) {
        return s + "\n";
    }
    return s;
}
//# sourceMappingURL=index.js.map